# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

from copy import deepcopy
from embodichain.lab.gym.utils import registration as env_registry
from embodichain.lab.gym.envs.embodied_env import EmbodiedEnvCfg


def build_env(env_id: str, base_env_cfg: EmbodiedEnvCfg):
    """Create env from registry id, auto-inferring cfg class (EnvName -> EnvNameCfg)."""
    env = env_registry.make(env_id, cfg=deepcopy(base_env_cfg))
    return env


__all__ = [
    "build_env",
]
